/*
 * Copyright (c) 1998, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "gameengine.h"

GameEngine::GameEngine()
{
	sounds = true;
	map = 0;
}

GameEngine::~GameEngine()
{
	deleteMonsters();
	delete [] map;
}

void GameEngine::newGame(Skill skill)
{
	game_skill = skill;
	xpos = ypos = 0;
	dir = South;
	sprite_wall = (Sprite)(Sprite_Wall1 + rand() % num_walls);
	sprite_back = (Sprite)(Sprite_Back1 + rand() % num_backs);
	deleteMonsters();
	status = Status_Playing;
	paused = false;

	int map_prop = 6;
	switch (skill) {
	case Skill_Beginner:
		map_prop = 6;
		break;
	case Skill_Intermediate:
		map_prop = 9;
		break;
	case Skill_Advanced:
		map_prop = 12;
		break;
	case Skill_Expert:
		map_prop = 15;
		break;
	}
	map_height = map_width = (4 * map_prop) + 1 - 4;
	if (map) delete [] map;
	map = new Sprite[map_width * map_height];

	for (int y = 0; y < map_height; ++y)
		for (int x = 0; x < map_width; ++x)
			ref(x, y) = sprite_wall;
	for (int y = 0; y < map_height; ++y)
		ref(0, y) = ref(map_width - 1, y) = sprite_back;
	for (int x = 0; x < map_width; ++x)
		ref(x, 0) = ref(x, map_height - 1) = sprite_back;

	// Compute the maze
	createMap(map_width / 2, map_height / 2);

	for (int x = 0; x < map_width; ++x)
		ref(x, 0) = ref(x, map_height - 1) = sprite_wall;
	for (int y = 0; y < map_height; ++y)
		ref(0, y) = ref(map_width - 1, y) = sprite_wall;

	// Clear the start and end path
	int pos[][2] = {{0, 0}, {0, 1}, {0, 2}, {1, 2}};
	for (int i = 0; i < 4; ++i)
		ref(pos[i][0], pos[i][1]) =
			ref(map_width - 1 - pos[i][0],
			    map_height - 1 - pos[i][1]) = sprite_back;

	// Place randomly the points
	num_pills = 25 * (skill + 1);
	for (int i = 0; i < num_pills; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (!(x == 0 && y == 0) && ref(x, y) == sprite_back) {
			ref(x, y) = Sprite_Pill;
			++i;
		}
	}

	// Remove randomly some walls
	for (int i = 0; i < (map_width * map_height) / 40; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (x >= 2 && y >= 2
		    && x <= map_width - 3 && y <= map_height - 3
		    && ref(x, y) == sprite_wall
		    && ((ref(x-1, y) == sprite_back && ref(x, y-1) == sprite_wall
			 && ref(x+1, y) == sprite_back && ref(x, y+1) == sprite_wall)
			|| (ref(x-1, y) == sprite_wall && ref(x, y-1) == sprite_back
			    && ref(x+1, y) == sprite_wall && ref(x, y+1) == sprite_back))) {
			ref(x, y) = sprite_back;
			++i;
		}
	}

	// Place randomly the monsters
	for (int i = 0; i < (skill + 1) * 4; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (x >= 10 && y >= 10 && ref(x, y) == sprite_back) {
			addMonster(x, y);
			++i;
		}
	}

	// Place randomly the potions
	for (int i = 0; i < (skill + 1) * 2; ) {
		int x = rand() % map_width;
		int y = rand() % map_height;
		if (x >= 3 && y >= 3 && ref(x, y) == sprite_back) {
			ref(x, y) = Sprite_Potion;
			++i;
		}
	}

	game_time = 0;
	potion_time = 0;
	balloon_time = 0;
}

void GameEngine::addMonster(int x, int y)
{
 	Monster *mp = new Monster(x, y, Sprite_Monster_Up);
	monsters.push_back(mp);
}

Monster *GameEngine::findMonster(int x, int y)
{
	typedef vector<Monster *>::const_iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i) {
		Monster *mp = *i;
		if (!mp->dead && mp->xpos == x && mp->ypos == y)
			return mp;
	}
	return 0;
}

void GameEngine::killMonster(Monster *mp)
{
	mp->dead = true;
	balloon = Sprite_Balloon_Chomp;
	balloon_time = 1000;
	playSound(Sound_Killed_Monster);
}

void GameEngine::deleteMonsters()
{
	typedef vector<Monster *>::iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i)
		delete *i;
	monsters.clear();
}

// Try to look around for Zacman
bool GameEngine::lookMonster(Monster *mp)
{
	if (mp->ypos == ypos) {
		for (int x = mp->xpos - 1; x >= 0; --x)
			if (xy(x, mp->ypos) == sprite_wall)
				break;
			else if (xpos == x) {
				mp->dir = West;
				return true;
			}
		for (int x = mp->xpos + 1; x <= map_width - 1; ++x)
			if (xy(x, mp->ypos) == sprite_wall)
				break;
			else if (xpos == x) {
				mp->dir = East;
				return true;
			}
	}
	if (mp->xpos == xpos) {
		for (int y = mp->ypos - 1; y >= 0; --y)
			if (xy(mp->xpos, y) == sprite_wall)
				break;
			else if (ypos == y) {
				mp->dir = North;
				return true;
			}
		for (int y = mp->ypos + 1; y <= map_height - 1; ++y)
			if (xy(mp->xpos, y) == sprite_wall)
				break;
			else if (ypos == y) {
				mp->dir = South;
				return true;
			}
	}

	return false;
}

Direction GameEngine::monsterFindDirection(Monster *mp, Direction exclude)
{
	if (exclude != South && xy(mp->xpos, mp->ypos + 1) != sprite_wall)
		return South;
	else if (exclude != West && xy(mp->xpos - 1, mp->ypos) != sprite_wall)
		return West;
	else if (exclude != North && xy(mp->xpos, mp->ypos - 1) != sprite_wall)
		return North;
	else if (exclude != East && xy(mp->xpos + 1, mp->ypos) != sprite_wall)
		return East;

	switch (exclude) {
	case North: return South;
	case South: return North;
	case West: return East;
	case East: return West;
	}

	return North; // Never happens
}

// Update the monsters' status
void GameEngine::updateMonsters()
{
	typedef vector<Monster *>::const_iterator I;
	for (I i = monsters.begin(); i != monsters.end(); ++i) {
		Monster *m = *i;
		if (m->dead)
			continue;
		Sprite s;
		while ((s = (Sprite)(Sprite_Monster_Up + rand() % 4)) == m->sprite)
			;
		m->sprite = s;
		if (lookMonster(m)) {
			if (potion_time > 0) { // Invert direction
				switch (m->dir) {
				case North:
					m->dir = monsterFindDirection(m, North);
					break;
				case South:
					m->dir = monsterFindDirection(m, South);
					break;
				case West:
					m->dir = monsterFindDirection(m, West);
					break;
				case East:
					m->dir = monsterFindDirection(m, East);
				}
			} else {
				if (!m->saw) {
					playSound(Sound_Saw_Monster);
					m->saw = true;
				}
			}
		} else
			m->saw = false;
		if (m->dir == North && xy(m->xpos, m->ypos - 1) != sprite_wall &&
				!findMonster(m->xpos, m->ypos - 1)) {
			if (xpos == m->xpos && ypos == m->ypos - 1) {
				if (potion_time > 0)
					killMonster(m);
				else {
					balloon = Sprite_Balloon_Ouch;
					balloon_time = 1000;
					playSound(Sound_Died);
					status = Status_Died;
				}
			} else
				--m->ypos;
		} else if (m->dir == South && xy(m->xpos, m->ypos + 1) != sprite_wall &&
				!findMonster(m->xpos, m->ypos + 1)) {
			if (xpos == m->xpos && ypos == m->ypos + 1) {
				if (potion_time > 0)
					killMonster(m);
				else {
					balloon = Sprite_Balloon_Ouch;
					balloon_time = 1000;
					playSound(Sound_Died);
					status = Status_Died;
				}
			} else
				++m->ypos;
		} else if (m->dir == East && xy(m->xpos + 1, m->ypos) != sprite_wall &&
				!findMonster(m->xpos + 1, m->ypos)) {
			if (xpos == m->xpos + 1 && ypos == m->ypos) {
				if (potion_time > 0)
					killMonster(m);
				else {
					balloon = Sprite_Balloon_Ouch;
					balloon_time = 1000;
					playSound(Sound_Died);
					status = Status_Died;
				}
			} else
				++m->xpos;
		} else if (m->dir == West && xy(m->xpos - 1, m->ypos) != sprite_wall &&
				!findMonster(m->xpos - 1, m->ypos)) {
			if (xpos == m->xpos - 1 && ypos == m->ypos) {
				if (potion_time > 0)
					killMonster(m);
				else {
					balloon = Sprite_Balloon_Ouch;
					balloon_time = 1000;
					playSound(Sound_Died);
					status = Status_Died;
				}
			} else
				--m->xpos;
		} else {
			Direction d = (Direction)(rand() % 4);
			m->dir = d;
		}
	}
}

void GameEngine::createMap(int x, int y)
{
	int dir = 0;

	while (ref(x, y + 2) != sprite_back || ref(x, y - 2) != sprite_back
	    || ref(x + 2, y) != sprite_back || ref(x - 2, y) != sprite_back) {
		bool found = false;
		while (!found) {
			dir = rand() % 4;
			switch (dir) {
			case 0:
				if (ref(x, y + 2) == sprite_wall)
					found = true;
				break;
			case 1:
				if (ref(x, y - 2) == sprite_wall)
					found = true;
				break;
			case 2:
				if (ref(x + 2, y) == sprite_wall)
					found = true;
				break;
			case 3:
				if (ref(x - 2, y) == sprite_wall)
					found = true;
			}
		}
		ref(x, y) = sprite_back;
		switch (dir) {
		case 0:
			ref(x, y + 2) = ref(x, y + 1) = sprite_back;
			createMap(x, y + 2);
			break;
		case 1:
			ref(x, y - 2) = ref(x, y - 1) = sprite_back;
			createMap(x, y - 2);
			break;
		case 2:
			ref(x + 2, y) = ref(x + 1, y) = sprite_back;
			createMap(x + 2, y);
			break;
		case 3:
			ref(x - 2, y) = ref(x - 1, y) = sprite_back;
			createMap(x - 2, y);
		}
	}
}

void GameEngine::checkPosition(int x, int y)
{
	if (xy(x, y) == Sprite_Pill) {
		ref(x, y) = sprite_back;
		--num_pills;
		playSound(Sound_Got_Pill);
	} else if (xy(x, y) == Sprite_Potion) {
		ref(x, y) = sprite_back;
		potion_time += 10000;
		playSound(Sound_Got_Potion);
	}
	if (x == map_width - 1 && y == map_height - 1)
		if (!num_pills) {
			status = Status_Finished;
		} else {
			balloon_time = 2000;
			balloon = Sprite_Balloon_Exit;
			playSound(Sound_Cannot_Exit);
		}
}

void GameEngine::goUp()
{
	if (paused || status != Status_Playing)
		return;

	Monster *m;
	if ((m = findMonster(xpos, ypos - 1)) != 0)
		if (potion_time > 0)
			killMonster(m);
		else {
			balloon = Sprite_Balloon_Ouch;
			balloon_time = 1000;
			playSound(Sound_Died);
			status = Status_Died;
			return;
		}
	if (xy(xpos, ypos - 1) != sprite_wall)
		--ypos;
	checkPosition(xpos, ypos);
	dir = North;
}

void GameEngine::goDown()
{
	if (paused || status != Status_Playing)
		return;

	Monster *m;
	if ((m = findMonster(xpos, ypos + 1)) != 0)
		if (potion_time > 0)
			killMonster(m);
		else {
			balloon = Sprite_Balloon_Ouch;
			balloon_time = 1000;
			playSound(Sound_Died);
			status = Status_Died;
			return;
		}
	if (xy(xpos, ypos + 1) != sprite_wall)
		++ypos;
	checkPosition(xpos, ypos);
	dir = South;
}

void GameEngine::goLeft()
{
	if (paused || status != Status_Playing)
		return;

	Monster *m;
	if ((m = findMonster(xpos - 1, ypos)) != 0)
		if (potion_time > 0)
			killMonster(m);
		else {
			balloon = Sprite_Balloon_Ouch;
			balloon_time = 1000;
			playSound(Sound_Died);
			status = Status_Died;
			return;
		}
	if (xy(xpos - 1, ypos) != sprite_wall)
		--xpos;
	checkPosition(xpos, ypos);
	dir = West;
}

void GameEngine::goRight()
{
	if (paused || status != Status_Playing)
		return;

	Monster *m;
	if ((m = findMonster(xpos + 1, ypos)) != 0)
		if (potion_time > 0)
			killMonster(m);
		else {
			balloon = Sprite_Balloon_Ouch;
			balloon_time = 1000;
			playSound(Sound_Died);
			status = Status_Died;
			return;
		}
	if (xy(xpos + 1, ypos) != sprite_wall)
		++xpos;
	checkPosition(xpos, ypos);
	dir = East;
}

void GameEngine::timer(int msecs)
{
	if (!paused) {
		game_time += msecs;
		if (potion_time > 0) {
			potion_time -= msecs;
			if (potion_time <= 0) {
				potion_time = 0;
				playSound(Sound_Potion_Timeout);
			}
		}
		if (balloon_time > 0) {
			balloon_time -= msecs;
			if (balloon_time < 0)
				balloon_time = 0;
		}
		if (game_time % 300 == 0)
			updateMonsters();
	}
}

